/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: ImageStack.c,v 0.9.1.4 1991/12/20 14:22:57 leei Exp $";

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>

#include "GenImageP.h"
#include "ImageP.h"
#include "ImViewportP.h"

#include "FileDialog.h"
#include "readImage.h"
#include "proto.h"

#include <stdlib.h>
#include <math.h>

static void AddIMSActions __((Widget w));

static void updateStackList __((Widget w));
static void updateFileList __((Widget w));

Boolean
ImPushImageStack _IMPL((w, im), ImageWidget w _AND GenericImage im )
{
    ImageWidgetPart *ipart = &(w->image);
    ExternalParams xparams;

    if ( ipart->stack_total == ipart->stack_len ) {
	ipart->stack_total += 10;
	ipart->stack = (GenericDisplayParams *)
	  XtRealloc( (char *) ipart->stack, 
		     ipart->stack_total * sizeof(GenericDisplayParams) );
    }

    if ( ipart->stack_len ) {
	GenericImage old = ipart->stack[0]->generic.image;
	if ( ImWidth(old) != ImWidth(im) || ImHeight(old) != ImHeight(im) ) {
	    fprintf( stderr, "Image must be same size as stack.\n" );
	    return False;
	}
	ImUninstall(ipart->stack[0]);
    }

    if ( w->image.stack_len )
      bcopy(ipart->stack, ipart->stack + 1, 
	    sizeof(GenericDisplayParams) * (w->image.stack_len) );

    XParamsFromWidget(&xparams, (Widget) w);
    ipart->stack[0] = ImCreateParams(im, &xparams);
    ipart->stack_len++;
    ImInstall(ipart->stack[0]);

    if ( ipart->stack_len == 1 ) {
	XtVaSetValues( (Widget) w,
		      XtNimageWidth,  ImWidth(im),
		      XtNimageHeight, ImHeight(im),
		      NULL );
    }

    if ( XtIsRealized((Widget)w) )
      ImRedraw(w, False);
    else {
	Dimension im_width  = (Dimension) rint(ipart->mag_x * ipart->width);
	Dimension im_height = (Dimension) rint(ipart->mag_x * ipart->height);

	if ( w->core.width  != im_width  ) w->core.width  = im_width;
	if ( w->core.height != im_height ) w->core.height = im_height;
    }

    updateStackList(ipart->stack_editor);
    return True;
}

void
ImPopImageStack _IMPL((w, redraw, destroy),
		      ImageWidget w _AND Boolean redraw _AND Boolean destroy )
{
    ImageWidgetPart *ipart = &(w->image);

    if ( ipart->stack_len ) {
	GenericDisplayParams disp = ipart->stack[0];
	GenericImage im;

	/* Actually pop image display */
	im = disp->generic.image;
	ImUninstall(disp); 
	bcopy( ipart->stack + 1, ipart->stack, 
	      sizeof(GenericDisplayParams) * (ipart->stack_len-1) );
	--ipart->stack_len;
	ipart->stack[ipart->stack_len] = 0;

	/* If there's no more images then make sure has some size. */
	if ( !ipart->stack_len ) {
	    ipart->width = ipart->height = 1;
	} else {
	    ImInstall(ipart->stack[0]);
	}

	/* Trash the display and image(?) */
	ImDestroyParams(disp, True);

	if ( redraw ) ImRedraw(w, False);
	updateStackList(ipart->stack_editor);
    }
}

void
ImRotImageStack _IMPL((w, n, redraw), 
		      ImageWidget w _AND int n _AND Boolean redraw )
{
    ImageWidgetPart *ipart = &(w->image);
    
    if ( ipart->stack_len ) {
	GenericDisplayParams move;

	ImUninstall(ipart->stack[0]);
	if ( n > 0 ) {
	    while ( n-- ) {
		move = ipart->stack[0];
		bcopy( ipart->stack + 1, ipart->stack, 
		       sizeof(GenericDisplayParams) * (ipart->stack_len-1) );
		ipart->stack[ipart->stack_len-1] = move;
	    }
	} else {
	    while ( n++ ) {
		move = ipart->stack[ipart->stack_len-1];
		bcopy( ipart->stack, ipart->stack + 1, 
		       sizeof(GenericDisplayParams) * (ipart->stack_len-1) );
		ipart->stack[0] = move;
	    }
	}
	ImInstall(ipart->stack[0]);
	
	if ( redraw ) ImRedraw(w, False);
	updateStackList(ipart->stack_editor);
    }
}

#include <X11/Shell.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/MenuButton.h>

#include <vim/EzMenu.h>
#include <vim/ScrollList.h>

static void imsRedrawCB __((Widget w, ImageWidget imw));
static void imsDelayRedrawsCB __((Widget w, ImageWidget imw));

Widget
ImCreateStackEditor _IMPL((imw), Widget imw )
{
    Widget shell, form, stBox, flBox, button;

    shell = XtVaCreatePopupShell("stackEditor", transientShellWidgetClass, imw,
				 XtNtransientFor, imw,
				 XtNallowShellResize, True,
				 NULL);
    XtAddCallback(shell, XtNpopupCallback, imsDelayRedrawsCB, (XtPointer)imw);
    XtAddCallback(shell, XtNpopdownCallback, imsRedrawCB, (XtPointer)imw);

    form = XtVaCreateManagedWidget("form", formWidgetClass, shell, NULL);

    stBox = XtVaCreateManagedWidget("stackBox", formWidgetClass, form, NULL);
    button =
      XtVaCreateManagedWidget("stackButton", menuButtonWidgetClass, stBox, 
			      XtNmenuName, "stackMenu", NULL);
    XtVaCreatePopupShell("stackMenu", ezMenuWidgetClass, button, NULL);
    XtVaCreateManagedWidget("stackList", ukcScrListWidgetClass, stBox, NULL);

    flBox = XtVaCreateManagedWidget("fileBox", formWidgetClass, form, NULL);
    button =
      XtVaCreateManagedWidget("fileButton", menuButtonWidgetClass, flBox, 
			      XtNmenuName, "fileMenu", NULL);
    XtVaCreatePopupShell("fileMenu", ezMenuWidgetClass, button, NULL);
    XtVaCreateManagedWidget("fileList", ukcScrListWidgetClass, flBox, NULL);

    {
	Widget apply =
	  XtVaCreateManagedWidget("apply", commandWidgetClass, form, NULL);
	XtAddCallback(apply, XtNcallback, imsRedrawCB, (XtPointer)imw);
	XtAddCallback(apply, XtNcallback, imsDelayRedrawsCB, (XtPointer)imw);
    }

    {
	Widget dismiss =
	  XtVaCreateManagedWidget("dismiss", commandWidgetClass, form, NULL);
	XtPopdownID popdown = XtNew(XtPopdownIDRec);
	popdown->enable_widget = imw;
	popdown->shell_widget = shell;
	XtAddCallback(dismiss, XtNcallback, XtCallbackPopdown, popdown);
    }

    AddIMSActions(shell);
    XtInstallAcceleratorsBelow(shell, shell);

    return shell;
}

static void
imsRedrawCB _IMPL((w, imw), Widget w _AND ImageWidget imw)
{
    ImRedraw(imw, True);
}

static void
imsDelayRedrawsCB _IMPL((w, imw), Widget w _AND ImageWidget imw)
{
    ImDelayRedraws(imw);
}

void
ImUseStackEditor _IMPL((w), Widget w )
{
    Boolean placed = XtIsRealized(w);

    updateStackList(w);
    updateFileList(w);

    XtRealizeWidget(w);
    PlacePopup(w, placed);
    XtPopup(w, XtGrabNone);
}

static Widget
NameToWidgetAbove _IMPL((ref, name), Widget ref _AND String name )
{
    for ( ; ref; ref = XtParent(ref) )
      if ( !strcmp(XtName(ref), name) )
	break;

    return ref;
}

static void
updateStackList _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    ImageWidget imw;
    Widget listw;
    GenericDisplayParams *sp;
    int i, len;
    char **list, **lp;

    if ( !st_editor ) return;
    imw = (ImageWidget) XtParent(st_editor);
    sp = imw->image.stack;
    len = imw->image.stack_len;

    listw = XtNameToWidget(st_editor, "form.stackBox.stackList");
    if ( !listw ) return;

    if ( len ) {
	/* Collect list of idents. */
	list = (char **) XtMalloc((len+1) * sizeof(char *));
	for ( i = 0, lp = list; i < len; ++i, ++lp ) {
	    *lp = (char *) ImIdent(sp[i]->generic.image);
	}
	*lp = 0;

	/* Initialize new list. */
	XukcScrListChange(listw, list, len);
    }
}

static int
selectedStackElem _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    Widget scr_listw = XtNameToWidget(st_editor, "form.stackBox.stackList");
    Widget listw = XtNameToWidget(scr_listw, "scrList");
    String *list;
    XawListReturnStruct *sel;
    int i, j, len;

    if ( !listw ) return;
    
    sel = XawListShowCurrent(listw);
    if ( !sel ) return -1;

    return sel->list_index;
}

static void
updateFileList _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    ImageWidget imw;
    Widget listw;
    int i;
    char **list;

    if ( !st_editor ) return;
    imw = (ImageWidget) XtParent(st_editor);

    listw = XtNameToWidget(st_editor, "form.fileBox.fileList");
    if ( !listw ) return;

    /* Collect list of filenames. */
    list = (char **) XtMalloc( (imFileListLength+1) * sizeof(char *));
    for ( i = 0; i < imFileListLength; ++i ) {
	char *p = strrchr(imFileList[i].filename, '/');
	list[i] = p? p+1: imFileList[i].filename;
    }
    list[i] = 0;

    /* Initialize new list. */
    XukcScrListChange(listw, list, imFileListLength);
}

static GenericImage
selectedFile _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    Widget scr_listw = XtNameToWidget(st_editor, "form.fileBox.fileList");
    Widget listw = XtNameToWidget(scr_listw, "scrList");
    String *list;
    XawListReturnStruct *sel;
    int i, j, len;

    if ( !listw ) return 0;
    
    sel = XawListShowCurrent(listw);
    if ( !sel ) return 0;

    len = strlen(sel->string);
    for ( j = 0; j < imFileListLength; ++j ) {
	char *filename = imFileList[j].filename;
	if ( !strcmp(sel->string, filename+strlen(filename)-len) )
	  return imFileList[j].image;
    }
    return 0;
}

static void
imsPush _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    Widget scr_listw = XtNameToWidget(st_editor, "form.stackBox.stackList");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    GenericImage im;

    im = selectedFile(st_editor);
    if ( !im ) return;

    ImPushImageStack(imw, im);
    updateStackList(st_editor);
    XukcScrListSelectItem(scr_listw, 1, True, True);
}

static void
imsPop _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    Widget scr_listw = XtNameToWidget(st_editor, "form.stackBox.stackList");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);

    ImPopImageStack(imw, False, False);
    updateStackList(st_editor);
    XukcScrListSelectItem(scr_listw, 1, True, True);
}

static void
imsRotate _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    Widget scr_listw = XtNameToWidget(st_editor, "form.stackBox.stackList");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    int idx;

    idx = selectedStackElem(st_editor);
    ImRotImageStack(imw, 1, False);
    updateStackList(st_editor);
    XukcScrListSelectItem(scr_listw, idx, True, True);
}

static void
imsDelete _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    Widget scr_listw = XtNameToWidget(st_editor, "form.stackBox.stackList");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    GenericDisplayParams *stack = imw->image.stack;
    int last = imw->image.stack_len;
    int i, idx;

    idx = selectedStackElem(st_editor);
    if ( idx < 0 ) return;

    ImUninstall(stack[0]);
    for ( i = idx; i < last-1; ++i )
      stack[i] = stack[i+1];
    --imw->image.stack_len;
    ImInstall(stack[0]);

    updateStackList(st_editor);
    XukcScrListSelectItem(scr_listw, idx+1, True, True);
}

static void
imsToTop _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    Widget scr_listw = XtNameToWidget(st_editor, "form.stackBox.stackList");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    GenericDisplayParams *stack = imw->image.stack;
    int i, idx;

    idx = selectedStackElem(st_editor);
    if ( idx < 0 ) return;

    ImUninstall(stack[0]);
    for ( i = idx; i > 0; --i ) {
	GenericDisplayParams temp = stack[i];
	stack[i] = stack[i-1];
	stack[i-1] = temp;
    }
    ImInstall(stack[0]);

    updateStackList(st_editor);
    XukcScrListSelectItem(scr_listw, 1, True, True);
}

static void
imsToBottom _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    Widget scr_listw = XtNameToWidget(st_editor, "form.stackBox.stackList");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    GenericDisplayParams *stack = imw->image.stack;
    int last = imw->image.stack_len;
    int i, idx;

    idx = selectedStackElem(st_editor);
    if ( idx < 0 ) return;

    ImUninstall(stack[0]);
    for ( i = idx; i < last-1; ++i ) {
	GenericDisplayParams temp = stack[i];
	stack[i] = stack[i+1];
	stack[i+1] = temp;
    }
    ImInstall(stack[0]);

    updateStackList(st_editor);
    XukcScrListSelectItem(scr_listw, last, True, True);
}

static void
imsEditParams _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    int idx;
    GenericDisplayParams disp;

    idx = selectedStackElem(st_editor);
    if ( idx < 0 ) return;

    disp = imw->image.stack[idx];
    if ( ImEditParams(disp) )
      ImRedraw(imw, False);
}

static void
imsToggleVisible _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    int idx;
    GenericDisplayParams disp;

    idx = selectedStackElem(st_editor);
    if ( idx < 0 ) return;

    disp = imw->image.stack[idx];
    disp->generic.visible = !disp->generic.visible;
    ImRedraw(imw, False);
}

static void
imsDone _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");

    if ( !st_editor ) return;
    XtPopdown(st_editor);
}

extern Widget image_fdlog;

static void
imsReadFile _IMPL((w), Widget w)
{
    Widget st_editor = NameToWidgetAbove(w, "stackEditor");
    Widget scr_listw = XtNameToWidget(st_editor, "form.fileBox.fileList");
    ImageWidget imw = (ImageWidget) XtParent(st_editor);
    char *fname;
    int idx;

    if ( !image_fdlog ) {
	image_fdlog = makeFileDialog("file", (Widget)findApplicationShell(w));
	setFileDialogAcceptFunc(image_fdlog, acceptImageFile);
    }

    fname = useFileDialog(image_fdlog, False);
    if ( !fname ) return;

    readImage(fname);

    updateFileList(st_editor);
    idx = ImFindFile(fname) - imFileList;
    XukcScrListSelectItem(scr_listw, idx+1, True, True);
    selectedFile(st_editor);
}

static XtActionsRec actions[] =
{
    /* {name, procedure}, */
    {"imsPush",		imsPush},
    {"imsPop",		imsPop},
    {"imsRotate",	imsRotate},
    {"imsDelete",	imsDelete},
    {"imsToTop",	imsToTop},
    {"imsToBottom",	imsToBottom},

    {"imsEditParams",	imsEditParams},
    {"imsToggleVisible",imsToggleVisible},

    {"imsReadFile",	imsReadFile},
    {"imsDone",		imsDone}
};

static void
AddIMSActions _IMPL((w), Widget w)
{
    XtAppAddActions( XtWidgetToApplicationContext(w), 
		     actions, XtNumber(actions) );
}
