/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: RGBImage.c,v 0.9.1.2 1991/12/09 19:36:55 leei Exp $";

#include <stdlib.h>
#include <math.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "RGBImageP.h"
#include <vim/CMManage.h>
#include <vim/gray.h>
#include <vim/gamma.h>

#include <vim/debug.h>
#include <vim/proto.h>

extern double rint __((double));

static XtResource resources[] = {
    { ImNdepth, ImCDepth, XtRShort, sizeof(short),
         XtOffset(RGBImage,rgb.depth), XtRImmediate, (XtPointer) 8},
    { ImNdata, ImCRGBData, ImRRGBData, sizeof(XtPointer),
         XtOffset(RGBImage,rgb.data), XtRImmediate, (XtPointer) 0}
};

static void
  Initialize __(( GenericImage im, ArgList argv, Cardinal argc ));
static void
  Destroy __(( GenericImage im ));

static void
  InitializeParams __(( GenericDisplayParams disp ));
static Boolean
  SetParams __(( GenericDisplayParams old, GenericDisplayParams req, 
		GenericDisplayParams disp));
static void
  EditParams __(( GenericDisplayParams disp ));
static void
  DescribeElem __(( GenericImage im, int i, int j, char *buf, int len ));

static int
  AllocColors __(( GenericDisplayParams disp ));
static void
  FreeColors __(( GenericDisplayParams disp ));
static void
  Redisplay __(( GenericDisplayParams disp, Boolean same ));

RGBImageClassRec rgbImageClassRec = {
  { /* core fields */
    /* superclass		*/	&genericImageClassRec,
    /* class_name		*/	"RGBImage",
    /* image_size		*/	sizeof(RGBImageRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* resources                */	resources,
    /* num_resources            */	XtNumber(resources),
    /* set_values               */      NULL,
    /* read			*/	NULL,
    /* write			*/	NULL,
    /* image_display_size	*/	sizeof(RGBDisplayParamsRec),
    /* param_resources          */	NULL,
    /* num_param_resources      */	0,
    /* initialize_params	*/	InitializeParams,
    /* set_params		*/	SetParams,
    /* destroy_params		*/	NULL,
    /* edit_params		*/	EditParams,
    /* remember_params		*/	NULL,
    /* compare_params		*/	NULL,
    /* install			*/	NULL,
    /* uninstall		*/	NULL,
    /* alloc_colors             */      AllocColors,
    /* free_colors              */      FreeColors,
    /* display			*/	Redisplay,
    /* description		*/	NULL,
    /* element_description	*/	DescribeElem,
    /* destroy			*/      Destroy
  }
};

ImageClass rgbImageClass = (ImageClass) &rgbImageClassRec;

static void
  Initialize _IMPL((gim, argv, argc),
		   GenericImage gim _AND ArgList argv _AND Cardinal argc )
{
    RGBImage im = (RGBImage) gim;
    int width  = im->generic.width;
    int height = im->generic.height;
    int depth  = im->rgb.depth;

    int pixlen;

    if ( depth == 0 ) im->rgb.depth = depth = 8;

    im->rgb.pixlen = pixlen = (depth + 7) / 8;

    if ( !im->rgb.data )
      im->rgb.data   = (byte *) XtCalloc(3 * width * height, pixlen);
}

static void
  Destroy _IMPL((gim), GenericImage gim )
{
    RGBImage im = (RGBImage) gim;

    XtFree((char *) im->rgb.data);
}

static void
  InitializeParams _IMPL((gdisp), GenericDisplayParams gdisp )
{
    RGBDisplayParams disp = (RGBDisplayParams) gdisp;
    ExternalParams *xparams = ImExternalParams(gdisp);

    if ( xparams->screen && DefaultDepthOfScreen(xparams->screen) == 1 )
      chooseGamma(&disp->rgb.gamma, gamEnhance);
    else
      defaultGamma(&disp->rgb.gamma);
    disp->rgb.redo_gamma = True;
}

static void
  DescribeElem _IMPL((gim, i, j, buf, len),
		     GenericImage gim _AND int i _AND int j _AND
		     char *buf _AND int len )
{
    RGBImage im = (RGBImage) gim;
    RGBImagePart *ipart = &im->rgb;
    int idx = 3*(j*im->generic.width + i);

    sprintf(buf, "%s: Value { %.2f %.2f %.2f }\n", ImIdent(gim), 
	    ipart->data[idx+0]/(float)MaxIntensity,
	    ipart->data[idx+1]/(float)MaxIntensity,
	    ipart->data[idx+2]/(float)MaxIntensity);
}

static Boolean
  SetParams _IMPL((g_old, g_req, g_disp),
		  GenericDisplayParams g_old _AND
		  GenericDisplayParams g_req _AND 
		  GenericDisplayParams g_disp) 
{
    RGBDisplayParams old  = (RGBDisplayParams) g_old;
    RGBDisplayParams req  = (RGBDisplayParams) g_req;
    RGBDisplayParams disp = (RGBDisplayParams) g_disp;
    
    return disp->rgb.redo_gamma;
}

static void
  EditParams _IMPL((gdisp), GenericDisplayParams gdisp )
{
    RGBDisplayParams disp = (RGBDisplayParams) gdisp;
    ExternalParams *xparams = ImExternalParams(gdisp);

    disp->rgb.redo_gamma =
      EditGamma(xparams->widget, &disp->rgb.gamma);
}

static void
  reduceRGBMap _DCL((disp), RGBDisplayParams disp);

static int
  AllocColors _IMPL((gdisp), GenericDisplayParams gdisp )
{
    RGBDisplayParams disp = (RGBDisplayParams) gdisp;
    RGBImage rgb = (RGBImage) disp->generic.image;
    ExternalParams *xparams = ImExternalParams(gdisp);
    use_RGBImage(rgb,rgb);
    Visual *vis = xparams->visual;
    GammaParams *gamma = &disp->rgb.gamma;
    int i, nalloc;
    XColor *xcolor, *colors;

    /* Allocate colors (only if a dynamic colormap is involved). */
    if ( vis->class != PseudoColor || 
	 (!disp->rgb.redo_gamma &&
	  disp->rgb.color_timestamp > rgb->generic.timestamp) ) 
      return 1;

    /* This is the actual meat of the 24->? reduction. */
    reduceRGBMap(disp);

    if ( rgb_cmap->used ) {
	if ( !disp->rgb.index )
	  disp->rgb.index = (Pixel *) malloc(sizeof(Pixel) * rgb_cmap->used);
	
	colors = (XColor *) alloca(sizeof(XColor) * rgb_cmap->used);
	
	/* Initialize the color list. */
	for ( i = 0, xcolor = colors; i < rgb_cmap->used; i++, ++xcolor ) {
	    xcolor->flags = DoRed | DoGreen | DoBlue;
	    xcolor->red   = rgb_cmap->red[i];
	    xcolor->green = rgb_cmap->green[i];
	    xcolor->blue  = rgb_cmap->blue[i];
	    gammaCorrect(gamma, xcolor);
	}
	
	if ( disp->rgb.color_timestamp > 0 )
	  CMFreeColors( disp->rgb.index, rgb_cmap->used );

	/* Now try to get all colors exactly. */
	nalloc = 
	  CMAllocColors(colors, rgb_cmap->used, /*exact*/True, /*all*/True);
	
	if ( nalloc != rgb_cmap->used ) {
	    /* For the moment, this is the cheap way out.  Eventually, 
	     * I'll back off and dither.  Now I just allocate as many as I 
	     * can and then use closest.  ** UPDATE **
	     */
	    nalloc = CMAllocColors(colors, rgb_cmap->used, False, True);
	}
	
	/*
	 * Now update the pixel entries.
	 */
	for ( i = 0, xcolor = colors; i < rgb_cmap->used; i++, ++xcolor )
	  disp->rgb.index[i] = xcolor->pixel;
    }

    time(&(disp->rgb.color_timestamp));
    return 1;
}

static void
FreeColors _IMPL((gdisp), GenericDisplayParams gdisp )
{
    RGBDisplayParams disp = (RGBDisplayParams) gdisp;
    RGBImage rgb = (RGBImage) disp->generic.image;
    use_RGBImage(rgb,rgb);

    if ( disp->rgb.color_timestamp > 0 )
      CMFreeColors( disp->rgb.index, rgb_cmap->used );
    if ( disp->rgb.index ) {
	free(disp->rgb.index);
	disp->rgb.index = 0;
    }
    disp->rgb.color_timestamp = 0;
}

static byte *
  pixelateImage _IMPL((disp, im, im_width, x, y, width, height, depth, llen),
		      RGBDisplayParams disp _AND 
		      byte *im _AND int im_width _AND
		      int x _AND int y _AND int width _AND int height _AND
		      int depth _AND int *llen )
{
    int i, j;
    int bytes = (depth+7)/8;
    byte *dp = &im[3*im_width*y], *p;
    int numpixels = 3 * width * height;
    byte *bp, *bdata = (byte *) malloc( sizeof(byte) * numpixels );

    *llen = 3 * bytes * width;
    for ( j=0, bp=bdata; j < height; ++j, dp += 3 * im_width )
      for ( i=0, p=dp+x ; i < width; ++i ) {
	  *bp++ = *p++; *bp++ = *p++; *bp++ = *p++;
      }
    return bdata;
}

static byte *
  zoomRGBImage _DCL((im, xmag, ymag, xp, yp, widthp, heightp),
		    RGBImage im _AND float xmag _AND float ymag _AND
		    int *xp _AND int *yp _AND int *widthp _AND int *heightp );

static byte *
  staticGrayDither _DCL((disp, im, im_width, x, y, width, height, depth, llen),
			RGBDisplayParams disp _AND 
			byte *im _AND int im_width _AND
			int x _AND int y _AND int width _AND int height _AND 
			int depth _AND int *llen);

static byte *
  trueColorDither _DCL((disp, im, im_width, x, y, width, height, depth, llen),
		       RGBDisplayParams disp _AND 
		       byte *im _AND int im_width _AND
		       int x _AND int y _AND int width _AND int height _AND
		       int depth _AND int *llen);

static byte *
  reduceImage _DCL((disp, im, im_width, sx, sy, width, height, depth, llen),
		   RGBDisplayParams disp _AND byte *im _AND int im_width _AND
		   int sx _AND int sy _AND int width _AND int height _AND 
		   int depth _AND int *llen );

static void
  Redisplay _IMPL((gdisp, same), GenericDisplayParams gdisp _AND Boolean same )
{
    RGBDisplayParams disp = (RGBDisplayParams) gdisp;
    RGBImage rgb = (RGBImage) disp->generic.image;
    ExternalParams *xparams = ImExternalParams(gdisp);
    Display *display = xparams->display;
    Window win = xparams->win;
    Visual *vis = xparams->visual;
    Screen *scrn = xparams->screen;
    float mag_x = xparams->mag_x;
    float mag_y = xparams->mag_y;
    Region region = xparams->region;
    XRectangle *range = &xparams->range;
    int ddepth = DefaultDepthOfScreen(scrn);
    int llen;
    int src_x, src_y, dst_x, dst_y, width, height, zim_width;
    Pixmap pm = rgb->rgb.pm;
    XImage *xim = 0;
    byte *zim = 0, *zim1 = 0;
    GC gc = 0; XGCValues gcv;
    use_RGBImage(rgb,rgb);
    
    dst_x = (int) rint(mag_x*range->x);
    dst_y = (int) rint(mag_y*range->y);
    width  = rgb->rgb.pm_width;
    height = rgb->rgb.pm_height;

    DEBUG((DBG_DISP, "Display RGB image %#x (%dx%d) at %d,%d\n", 
	   rgb, rgb_width, rgb_height, dst_x, dst_y));

    if ( !same || !pm || disp->rgb.redo_gamma ) {
	int format;
	
	/* Free the previous pixmap. */
	if ( pm ) XFreePixmap(display,pm);
	
	/* Create an image on the given range. */
	width  = range->width;
	height = range->height;
	
	if ( mag_x == 1.0 && mag_y == 1.0 ) {
	    zim = rgb_data;
	    zim_width = rgb_width;
	    src_x = dst_x; src_y = dst_y;
	} else {
	    src_x = range->x; src_y = range->y;
	    zim = zoomRGBImage(rgb, mag_x,mag_y, 
			       &src_x,&src_y, &width,&height);
	    DEBUG((DBG_DISP,"Zoom image %#x at %d,%d by %g,%g => %#x (%dx%d)\n", 
		   rgb, src_x, src_y, mag_x, mag_y, zim, width, height));
	    if ( !zim ) goto cleanup;
	    zim_width = width;
	    src_x = src_y = 0;
	}
	
	switch ( vis->class ) {
	  case StaticColor:
	  case TrueColor:
	    zim = trueColorDither(disp, zim1=zim, zim_width,
				  src_x,src_y, width,height,ddepth, &llen);
	    DEBUG((DBG_DISP, "TC dither image %#x (%dx%d) at %d,%d => %#x\n", 
		   zim1, rgb_width, rgb_height, src_x, src_y, zim));
	    if ( zim1 != rgb_data ) free(zim1);
	    zim1 = 0;
	    format = ZPixmap;
	    src_x = src_y = 0;
	    break;
	  case GrayScale:
	  case StaticGray:
	    zim = staticGrayDither(disp, zim1=zim, zim_width,
				   src_x,src_y, width,height,ddepth, &llen);
	    DEBUG((DBG_DISP, "SG dither image %#x (%dx%d) at %d,%d => %#x\n", 
		   zim1, rgb_width, rgb_height, src_x, src_y, zim));
	    if ( zim1 != rgb_data ) free(zim1);
	    zim1 = 0;
	    format = ddepth==1? XYBitmap: ZPixmap;
	    src_x = src_y = 0;
	    break;
	  case PseudoColor:
	    zim = reduceImage(disp, zim1=zim,zim_width,
			      src_x,src_y, width,height,ddepth, &llen);
	    DEBUG((DBG_DISP, "Reduce image %#x (%dx%d) at %d,%d => %#x\n", 
		   zim1, rgb_width, rgb_height, src_x, src_y, zim));
	    if ( zim1 != rgb_data ) free(zim1);
	    zim1 = 0;
	    format = ZPixmap;
	    src_x = src_y = 0;
	    break;
	  default:
	    fprintf(stderr,"Cannot handle visual type %d\n",vis->class);
	    goto cleanup;
	}
	if ( !zim ) goto cleanup;
	xim = XCreateImage(display, vis, ddepth, format,
			   0, (char*) zim, width, height, 8, llen );
	xim->bitmap_bit_order = MSBFirst;
	xim->byte_order = MSBFirst;
	DEBUG((DBG_DISP, "Create image from %#x (%dx%d) => %#x\n", 
	       zim, width, height, xim));
    }

    /* Create the appropriate GC. */
    gcv.function = GXcopy;
    gcv.foreground = CMwhitePixel;
    gcv.background = CMblackPixel;
    gc = XCreateGC(display, win, GCFunction|GCForeground|GCBackground, &gcv);
    
    if ( xim ) {
	/* Finally draw it. */
	rgb->rgb.pm = pm = XCreatePixmap(display, win, width, height, ddepth);
	rgb->rgb.pm_width = width;
	rgb->rgb.pm_height = height;
	DEBUG((DBG_DISP, "Put image %#x at %d,%d (%dx%d) into pixmap %#x\n", 
	       xim, src_x, src_y, width, height, pm));
	XPutImage(display, pm, gc, xim, src_x, src_y, 0, 0, width, height);
    }

    XSetRegion( display, gc, region ); /**/
    DEBUG((DBG_DISP, "CopyArea %#x (%dx%d) to %d,%d\n", 
	   pm, width, height, dst_x, dst_y));
    XCopyArea(display, pm, win, gc, 0, 0, width, height, dst_x, dst_y);
    
    /* And clean up. */
cleanup:
    disp->rgb.redo_gamma = False;
    if ( zim && zim != rgb_data ) free(zim);
    if ( xim ) free(xim);
    if ( gc )  XFreeGC(display,gc);
}

static unsigned int *
  buildIndex _IMPL((x, width, mag, rx, rwidth),
		   int x _AND int width _AND float mag _AND 
		   int *rx _AND int *rwidth)
{
    unsigned int *index;
    unsigned int  a;

    *rwidth = (int) rint(mag * width);
    *rx = 0;

    index = (unsigned int *) malloc(*rwidth * sizeof(unsigned int));

    for (a = *rx; a < *rwidth; a++)
      index[a] = a/mag + x ;
    return(index);
}

static byte *
zoomRGBImage _IMPL((im, xmag, ymag, xp, yp, widthp, heightp),
		   RGBImage im _AND float xmag _AND float ymag _AND
		   int *xp _AND int *yp _AND int *widthp _AND int *heightp )
{
    use_RGBImage(im,im);
    unsigned int *xindex, *yindex;
    unsigned int  x, y, xsrc, pxsrc, ysrc, pysrc;
    unsigned int  srclinelen;
    unsigned int  destlinelen;
    byte         *dest;
    byte         *srcline, *srcptr, *psrcline, *psrcptr;
    byte         *destline, *destptr;
    unsigned int  srcwidth  = *widthp;
    unsigned int  srcheight = *heightp;
    unsigned int  srcx = *xp;
    unsigned int  srcy = *yp;

    xindex = buildIndex(srcx, srcwidth,  xmag, xp, widthp);
    yindex = buildIndex(srcy, srcheight, ymag, yp, heightp);

    destlinelen = 3 * *widthp;
    dest = destline = (byte *) calloc( *heightp * destlinelen, 1 );

    srclinelen = 3 * im_width;
    srcline = im_data + srcy * srclinelen + 3 * srcx;
  
    if ( xmag < 1.0 || ymag < 1.0 ) {
	for (y = 0, ysrc = yindex[y]; y < *heightp; y++) {
	    pysrc    = ysrc;
	    psrcline = srcline;

	    ysrc     = yindex[y];
	    srcline += (ysrc-pysrc) * srclinelen;

	    srcptr   = srcline;
	    destptr  = destline;
	    
	    for (x = 0, xsrc = xindex[x]; x < *widthp; x++) {
		short n = 0;
		int srcr = 0, srcg = 0, srcb = 0;

		pxsrc    = xsrc;
		xsrc     = xindex[x];
		srcptr  += 3 * (xsrc-pxsrc);
		psrcptr  = srcptr;
		psrcline = srcline;

		do {
		    do {
			++pxsrc; ++n;
			srcr += psrcptr[0]; 
			srcg += psrcptr[1]; 
			srcb += psrcptr[2];
			psrcptr -= 3;
		    } while ( xsrc > pxsrc );
		    ++pysrc;
		    psrcline -= srclinelen;
		    psrcptr   = psrcline + 3*xsrc;
		} while ( ysrc > pysrc );

		*destptr++ = srcr/n;
		*destptr++ = srcg/n;
		*destptr++ = srcb/n;
	    }
	    destline += destlinelen;
	}
    } else {
	for (y = 0, ysrc = yindex[y]; y < *heightp; y++) {
	    pysrc    = ysrc;
	    psrcline = srcline;

	    ysrc     = yindex[y];
	    if ( ysrc > pysrc )
	      srcline += (ysrc-pysrc) * srclinelen;

	    srcptr   = srcline;
	    destptr  = destline;
	    
	    for (x = 0, xsrc = xindex[x]; x < *widthp; x++) {
		byte srcr, srcg, srcb;

		pxsrc   = xsrc;
		xsrc    = xindex[x];
		if ( xsrc > pxsrc )
		  srcptr += 3 * (xsrc-pxsrc);

		srcr = srcptr[0]; srcg = srcptr[1]; srcb = srcptr[2];

		*destptr++ = srcr;
		*destptr++ = srcg;
		*destptr++ = srcb;
	    }
	    destline += destlinelen;
	}
    }

    free((char *)xindex);
    free((char *)yindex);

    return dest;
}

static void 
  GLeftToRight __((short inc, short *curr, short *next, short width));
static void 
  GRightToLeft __((short inc, short *curr, short *next, short width));

static byte *
  staticGrayDither _IMPL((disp, im, im_width, 
			  x, y, width, height, depth, llen),
			 RGBDisplayParams disp _AND
			 byte *im _AND int im_width _AND
			 int x _AND int y _AND int width _AND int height _AND
			 int depth _AND int *llen)
{
    GammaParams *gamma = &disp->rgb.gamma;
    byte perpxl = 8/depth;
    short maxval = (1<<depth)-1;
    short incval = (MaxGrey+maxval)/maxval;
    short destshft;
    short destlinelen;
    byte *dest, *destptr, *destline;
    short srclinelen;
    byte *srcline, *sp;
    short *curr, *next, *swap;
    int i, j;

    if ( depth > 8 ) {
	fprintf(stderr, "Unable to dither for %d bit displays.\n", depth );
	free(im);
	return 0;
    }

    /* Setup circulating line buffers. */
    curr = (short *) alloca( sizeof(short) * (width+2) ); ++curr;
    next = (short *) alloca( sizeof(short) * (width+2) ); ++next;
    for ( i=0; i<width+2; ++i ) curr[i] = next[i] = 0;

    srclinelen = 3 * im_width;
    srcline = im + y*srclinelen + 3*x;

    *llen = destlinelen = (width+perpxl-1)/perpxl;
    dest = destline = (byte *) calloc( height * destlinelen, 1 );
    
    for (j = 0; j < height; j++ ) {
	destptr  = destline;
	destshft = (8-depth);
	  
	for ( i = 0, sp = srcline; i < width; i++ ) {
	    XColor xcolor;
	    xcolor.red   = byteIntens(*sp); sp++;
	    xcolor.green = byteIntens(*sp); sp++;
	    xcolor.blue  = byteIntens(*sp); sp++;
	    curr[i+1] = grayIntens(gammaCorrect(gamma, &xcolor));
	}

	/* Either FS dither from left->right or right->left (alternate). */
	if ( j & 1 )
	  GLeftToRight( incval, curr+1, next+1, width );
	else
	  GRightToLeft( incval, curr+1, next+1, width );
	  
	/* Pack the FS dithered values into the pixmap */
	if ( depth == 8 ) {
	    for ( i = 0; i < width; i++ )
	      *destptr++ = curr[i+1]/incval;
	} else {
	    for ( i = 0; i < width; i++ ) {
		unsigned short val = curr[i+1]/incval;
		if (val) *destptr |= (val<<destshft);
		if ( (destshft -= depth) < 0) {
		    destshft = (8-depth);
		    destptr++;
		}
	    }
	}

	srcline  += srclinelen;
	destline += destlinelen;

	/* swap curr and next */
	swap = curr; curr = next; next = swap;
	for ( i=0; i<width+2; ++i ) next[i] = 0;
    }

    return dest;
}


/*
 * dither a line from left to right
 */
static void 
  GLeftToRight _IMPL((inc, curr, next, width),
		     short inc _AND short *curr _AND short *next _AND 
		     short width)
{
    short error;
    short idx;
    unsigned short output;
    unsigned short off = inc/2;

    for ( idx = 0; idx < width; idx++ ) {
	output       = (curr[idx]+off)/inc;
	output      *= inc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx-1] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx+1] += error * 1 / 16;
	curr[idx+1] += error * 7 / 16;
    }
}


/*
 * dither a line from right to left
 */
static void 
  GRightToLeft _IMPL((inc, curr, next, width),
		     short inc _AND short *curr _AND short *next _AND
		     short width)
{
    short error;
    short idx;
    unsigned short output;
    unsigned short off = inc/2;

    for ( idx = (width-1); idx >= 0; idx-- ) {
	output       = (curr[idx]+off)/inc;
	output      *= inc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx+1] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx-1] += error * 1 / 16;
	curr[idx-1] += error * 7 / 16;
    }
}

static void 
  CLeftToRight __((short rinc, short ginc, short binc, 
		   short *curr, short *next, short width));
static void 
  CRightToLeft __((short rinc, short ginc, short binc, 
		   short *curr, short *next, short width));

static void
  decode_mask _IMPL((mask, max, inc, shft),
		    int mask _AND short *max _AND short *inc _AND short *shft)
{
    *shft = 0; *max = 1;
    while ( !(mask & 0x01) ) {
	mask >>= 1;
	++*shft;
    }
    while ( mask & 0x01 ) {
	mask >>= 1;
	*max <<= 1;
    }
    --*max;
    *inc = (MaxGrey + *max) / *max;
}

static byte *
  trueColorDither _IMPL((disp, im, im_width, x, y, width, height, depth, llen),
			RGBDisplayParams disp _AND
			byte *im _AND int im_width _AND 
			int x _AND int y _AND int width _AND int height _AND
			int depth _AND int *llen)
{
    GammaParams *gamma = &disp->rgb.gamma;
    ExternalParams *xparams = ImExternalParams((GenericDisplayParams)disp);
    Visual *vis = xparams->visual;
    byte perpxl = 8/depth;
    short rmax, rinc, rshft;
    short gmax, ginc, gshft;
    short bmax, binc, bshft;
    short destshft;
    short destlinelen;
    byte *dest, *destptr, *destline;
    short srclinelen;
    byte *srcline, *sp;
    short *curr, *next, *swap;
    int i, j;

    decode_mask( vis->red_mask,   &rmax, &rinc, &rshft );
    decode_mask( vis->green_mask, &gmax, &ginc, &gshft );
    decode_mask( vis->blue_mask,  &bmax, &binc, &bshft );
    
#define ENCODE(r,g,b)\
    (((r/rinc)<<rshft) + ((g/ginc)<<gshft) + ((b/binc)<<bshft))

    if ( depth > 8 ) {
	fprintf(stderr, "Unable to dither for %d bit displays.\n", depth );
	free(im);
	return 0;
    }

    /* Setup circulating line buffers. */
    curr = (short *) alloca( 3 * sizeof(short) * (width+2) ); ++curr;
    next = (short *) alloca( 3 * sizeof(short) * (width+2) ); ++next;
    for ( i=0; i<3*(width+2); ++i ) curr[i] = next[i] = 0;

    srclinelen = 3 * im_width;
    srcline = im + y*srclinelen + 3*x;

    *llen = destlinelen = (width+perpxl-1)/perpxl;
    dest = destline = (byte *) calloc( height * destlinelen, 1 );
	
    for (j = 0; j < height; j++ ) {
	destptr  = destline;
	destshft = (8-depth);
	  
	for (i = 3, sp = srcline; i < 3*(width+1); ) {
	    XColor xcolor;
	    xcolor.red   = byteIntens(*sp); *sp++;
	    xcolor.green = byteIntens(*sp); *sp++;
	    xcolor.blue  = byteIntens(*sp); *sp++;
	    gammaCorrect(gamma, &xcolor);
	    curr[i++] += intensGray(xcolor.red);
	    curr[i++] += intensGray(xcolor.green);
	    curr[i++] += intensGray(xcolor.blue);
	}

	/* Either FS dither from left->right or right->left (alternate). */
	if ( j & 1 )
	  CRightToLeft( rinc,ginc,binc, curr+3, next+3, width );
	else
	  CLeftToRight( rinc,ginc,binc, curr+3, next+3, width );
	
	/* Pack the FS dithered values into the pixmap */
	if ( depth == 8 ) {
	    for ( i = 3; i < 3*(width+1); i += 3 )
	      *destptr++ = ENCODE(curr[i], curr[i+1], curr[i+2]);
	} else {
	    for (i = 3; i < 3*(width+1); i += 3 ) {
		byte val = ENCODE(curr[i], curr[i+1], curr[i+2]);
		if (val) *destptr |= (val<<destshft);
		if ( (destshft -= depth) < 0) {
		    destshft = (8-depth);
		    destptr++;
		}
	    }
	}
	srcline  += srclinelen;
	destline += destlinelen;

	/* swap curr and next */
	swap = curr; curr = next; next = swap;
	for ( i=0; i<3*(width+2); ++i ) next[i] = 0;
    }
#undef ENCODE

    return dest;
}


/*
 * dither a line from left to right
 */
static void 
  CLeftToRight _IMPL((rinc, ginc, binc, curr, next, width),
		     short rinc _AND short ginc _AND short binc _AND 
		     short *curr _AND short *next _AND short width)
{
    short error;
    short idx;
    unsigned short output;
    unsigned short roff = rinc/2;
    unsigned short goff = ginc/2;
    unsigned short boff = binc/2;

    for ( idx = 0; idx < 3*width; ) {
	/* Red */
	output       = (curr[idx]+roff)/rinc;
	output      *= rinc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx-3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx+3] += error * 1 / 16;
	curr[idx+3] += error * 7 / 16;
	++idx;

	/* Green */
	output       = (curr[idx]+goff)/ginc;
	output      *= ginc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx-3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx+3] += error * 1 / 16;
	curr[idx+3] += error * 7 / 16;
	++idx;
	
	/* Blue */
	output       = (curr[idx]+boff)/binc;
	output      *= binc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx-3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx+3] += error * 1 / 16;
	curr[idx+3] += error * 7 / 16;
	++idx;
    }
}


/*
 * dither a line from right to left
 */
static void 
  CRightToLeft _IMPL((rinc, ginc, binc, curr, next, width),
		     short rinc _AND short ginc _AND short binc _AND 
		     short *curr _AND short *next _AND short width)
{
    short error;
    short idx;
    unsigned short output;
    unsigned short roff = rinc/2;
    unsigned short goff = ginc/2;
    unsigned short boff = binc/2;

    for ( idx = (3*width-1); idx >= 0; ) {
	/* Blue */
	output       = (curr[idx]+boff)/binc;
	output      *= binc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx+3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx-3] += error * 1 / 16;
	curr[idx-3] += error * 7 / 16;
	--idx;

	/* Green */
	output       = (curr[idx]+goff)/ginc;
	output      *= ginc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx+3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx-3] += error * 1 / 16;
	curr[idx-3] += error * 7 / 16;
	--idx;

	/* Red */
	output       = (curr[idx]+roff)/rinc;
	output      *= rinc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx+3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx-3] += error * 1 / 16;
	curr[idx-3] += error * 7 / 16;
	--idx;
    }
}

/* reduce.c:
 *
 * reduce an image's colormap usage to a set number of colors.  this also
 * translates a true color image to a TLA-style image of `n' colors.
 *
 * this uses an algorithm by Paul Heckbert discussed in `Color Image
 * Quantization for Frame Buffer Display,' _Computer Graphics_ 16(3),
 * pp 297-307.  this implementation is based on one discussed in
 * 'A Few Good Colors,' _Computer Language_, Aug. 1990, pp 32-41 by
 * Dave Pomerantz.
 *
 * this function cannot reduce to any number of colors larger than 32768.
 *
 * jim frost 04.18.91
 *
 * Copyright 1991 Jim Frost.
 * See included file "jfcopyright.h" for complete copyright information.
 */

#include "jfcopyright.h"

#define DIST(A, B) ((A) < (B) ? (B) - (A) : (A) - (B))

/* find the distance between two colors.  we loose some accuracy here because
 * a triple squared short may not fit in a long.  we use a table lookup
 * to help speed this up; it's an O(exp(n,2)) algorithm.
 */

unsigned int  squareInit= 0;
unsigned long squareTable[32768];

void 
initSquareTable _IMPL_VOID
{ 
    unsigned long a;

    for (a= 0; a < 32768; a++)
      squareTable[a]= a * a;
    squareInit= 1;
}

unsigned long
colorDistance _IMPL((rgb, a, b), 
		    RGBMap *rgb _AND Pixel a _AND Pixel b)
{
    return(squareTable[DIST(*(rgb->red + a), *(rgb->red + b)) >> 1] +
	   squareTable[DIST(*(rgb->green + a), *(rgb->green + b)) >> 1] +
	   squareTable[DIST(*(rgb->blue + a), *(rgb->blue + b)) >> 1]);
}

/* this converts a TLA-style pixel into a 15-bit true color pixel
 */

#define TLA_TO_15BIT(TABLE,PIXEL)           \
  ((((TABLE).red[PIXEL] & 0xf800) >> 1) |   \
   (((TABLE).green[PIXEL] & 0xf800) >> 6) | \
   (((TABLE).blue[PIXEL] & 0xf800) >> 11))

/* this converts a 24-bit true color pixel into a 15-bit true color pixel
 */

#define TRUE_TO_15BIT(PIXEL)     \
  ((((PIXEL) & 0xf80000) >> 9) | \
   (((PIXEL) & 0x00f800) >> 6) | \
   (((PIXEL) & 0x0000f8) >> 3))

#define TRUE3_TO_15BIT(R,G,B)     \
  ((((R) & 0xf8) <<  7) | \
   (((G) & 0xf8) <<  2) | \
   (((B) & 0xf8) >>  3))

/* these macros extract color intensities from a 15-bit true color pixel
 */

#define RED_INTENSITY(P)   (((P) & 0x7c00) >> 10)
#define GREEN_INTENSITY(P) (((P) & 0x03e0) >> 5)
#define BLUE_INTENSITY(P)   ((P) & 0x001f)

/* this structure defines a color area which is made up of an array of pixel
 * values and a count of the total number of image pixels represented by
 * the area.  color areas are kept in a list sorted by the number of image
 * pixels they represent.
 */

struct color_area {
    unsigned short    *pixels;       /* array of pixel values in this area */
    unsigned short     num_pixels;   /* size of above array */
    int              (*sort_func)(); /* predicate func to sort with before
				      * splitting */
    unsigned long      pixel_count;  /* # of image pixels we represent */
    struct color_area *prev, *next;
};

/* predicate functions for qsort
 */

static int 
sortRGB _IMPL((p1, p2),
	      unsigned short *p1 _AND unsigned short *p2)
{ 
    unsigned int red1, green1, blue1, red2, green2, blue2;
    
    red1= RED_INTENSITY(*p1);
    green1= GREEN_INTENSITY(*p1);
    blue1= BLUE_INTENSITY(*p1);
    red2= RED_INTENSITY(*p2);
    green2= GREEN_INTENSITY(*p2);
    blue2= BLUE_INTENSITY(*p2);
    
    if (red1 == red2)
      if (green1 == green2)
	if (blue1 < blue2)
	  return(-1);
	else
	  return(1);
      else if (green1 < green2)
	     return(-1);
      else
	return(1);
    else if (red1 < red2)
	   return(-1);
    else
      return(1);
}

static int 
sortRBG _IMPL((p1, p2), 
	      unsigned short *p1 _AND unsigned short *p2)
{
    unsigned int red1, green1, blue1, red2, green2, blue2;
    
    red1= RED_INTENSITY(*p1);
    green1= GREEN_INTENSITY(*p1);
    blue1= BLUE_INTENSITY(*p1);
    red2= RED_INTENSITY(*p2);
    green2= GREEN_INTENSITY(*p2);
    blue2= BLUE_INTENSITY(*p2);
    
    if (red1 == red2)
      if (blue1 == blue2)
	if (green1 < green2)
	  return(-1);
	else
	  return(1);
      else if (blue1 < blue2)
	     return(-1);
      else
	return(1);
    else if (red1 < red2)
	   return(-1);
    else
      return(1);
}

static int
 sortGRB _IMPL((p1, p2), 
	       unsigned short *p1 _AND unsigned short *p2)
{ 
    unsigned int red1, green1, blue1, red2, green2, blue2;
    
    red1= RED_INTENSITY(*p1);
    green1= GREEN_INTENSITY(*p1);
    blue1= BLUE_INTENSITY(*p1);
    red2= RED_INTENSITY(*p2);
    green2= GREEN_INTENSITY(*p2);
    blue2= BLUE_INTENSITY(*p2);
    
    if (green1 == green2)
      if (red1 == red2)
	if (blue1 < blue2)
	  return(-1);
	else
	  return(1);
      else if (red1 < red2)
	     return(-1);
      else
	return(1);
    else if (green1 < green2)
	   return(-1);
    else
      return(1);
}

static int 
sortGBR _IMPL((p1, p2),
	      unsigned short *p1 _AND unsigned short *p2)
{ 
    unsigned int red1, green1, blue1, red2, green2, blue2;
    
    red1= RED_INTENSITY(*p1);
    green1= GREEN_INTENSITY(*p1);
    blue1= BLUE_INTENSITY(*p1);
    red2= RED_INTENSITY(*p2);
    green2= GREEN_INTENSITY(*p2);
    blue2= BLUE_INTENSITY(*p2);
    
    if (green1 == green2)
      if (blue1 == blue2)
	if (red1 < red2)
	  return(-1);
	else
	  return(1);
      else if (blue1 < blue2)
	     return(-1);
      else
	return(1);
    else if (green1 < green2)
	   return(-1);
    else
      return(1);
}

static int 
sortBRG _IMPL((p1, p2), 
	      unsigned short *p1 _AND unsigned short *p2)
{
    unsigned int red1, green1, blue1, red2, green2, blue2;
    
    red1= RED_INTENSITY(*p1);
    green1= GREEN_INTENSITY(*p1);
    blue1= BLUE_INTENSITY(*p1);
    red2= RED_INTENSITY(*p2);
    green2= GREEN_INTENSITY(*p2);
    blue2= BLUE_INTENSITY(*p2);
    
    if (blue1 == blue2)
      if (red1 == red2)
	if (green1 < green2)
	  return(-1);
	else
	  return(1);
      else if (red1 < red2)
	     return(-1);
      else
	return(1);
    else if (blue1 < blue2)
	   return(-1);
    else
      return(1);
}

static int
 sortBGR _IMPL((p1, p2), 
	       unsigned short *p1 _AND unsigned short *p2)
{ 
    unsigned int red1, green1, blue1, red2, green2, blue2;
    
    red1= RED_INTENSITY(*p1);
    green1= GREEN_INTENSITY(*p1);
    blue1= BLUE_INTENSITY(*p1);
    red2= RED_INTENSITY(*p2);
    green2= GREEN_INTENSITY(*p2);
    blue2= BLUE_INTENSITY(*p2);
    
    if (blue1 == blue2)
      if (green1 == green2)
	if (red1 < red2)
	  return(-1);
	else
	  return(1);
      else if (green1 < green2)
	     return(-1);
      else
	return(1);
    else if (blue1 < blue2)
	   return(-1);
    else
      return(1);
}

/* this does calculations on a color area following a split and inserts
 * the color area in the list of color areas.
 */

static void 
insertColorArea _IMPL((pixel_counts, rlargest, rsmallest, area),
		      unsigned long *pixel_counts _AND
		      struct color_area **rlargest _AND
		      struct color_area **rsmallest _AND
		      struct color_area *area)
{ 
    int a;
    unsigned int red, green, blue;
    unsigned int min_red, min_green, min_blue;
    unsigned int max_red, max_green, max_blue= 0;
    struct color_area *largest, *smallest, *tmp_area;
    
    min_red= min_green= min_blue= 31;
    max_red= max_green= max_blue= 0;
    
    /* update pixel count for this area and find RGB intensity widths
     */
    
    area->pixel_count= 0;
    for (a= 0; a < area->num_pixels; a++) {
	area->pixel_count += pixel_counts[area->pixels[a]];
	red= RED_INTENSITY(area->pixels[a]);
	green= GREEN_INTENSITY(area->pixels[a]);
	blue= BLUE_INTENSITY(area->pixels[a]);
	if (red < min_red)
	  min_red= red;
	if (red > max_red)
	  max_red= red;
	if (green < min_green)
	  min_green= green;
	if (green > max_green)
	  max_green= green;
	if (blue < min_blue)
	  min_blue= blue;
	if (blue > max_blue)
	  max_blue= blue;
    }
    
    /* calculate widths and determine which predicate function to use based
     * on the result
     */
    
    red= max_red - min_red;
    green= max_green - min_green;
    blue= max_blue - min_blue;
    
    if (red > green)
      if (green > blue)
	area->sort_func= sortRGB;
      else if (red > blue)
	     area->sort_func= sortRBG;
      else
	area->sort_func= sortBRG;
    else if (green > blue)
	   if (red > blue)
	     area->sort_func= sortGRB;
	   else
	     area->sort_func= sortGBR;
    else
      area->sort_func= sortBGR;
    
    /* insert color area in color area list sorted by number of pixels that
     * the area represents
     */
    
    largest= *rlargest;
    smallest= *rsmallest;
    
    if (!largest) {
	largest= smallest= area;
	area->prev= area->next= (struct color_area *)NULL;
    }
    
    /* if we only have one element, our pixel count is immaterial so we get
     * stuck on the end of the list.
     */
    
    else if (area->num_pixels < 2) {
	smallest->next= area;
	area->prev= smallest;
	area->next= (struct color_area *)NULL;
	smallest= area;
    }
    
    /* insert node into list
     */
    
    else {
	for (tmp_area= largest; tmp_area; tmp_area= tmp_area->next)
	  if ((area->pixel_count > tmp_area->pixel_count) ||
	      (tmp_area->num_pixels < 2)) {
	      area->prev= tmp_area->prev;
	      area->next= tmp_area;
	      tmp_area->prev= area;
	      if (area->prev)
		area->prev->next= area;
	      else
		largest= area;
	      break;
	  }
	if (!tmp_area) {
	    area->prev= smallest;
	    area->next= (struct color_area *)NULL;
	    smallest->next= area;
	    smallest= area;
	}
    }
    *rlargest= largest;
    *rsmallest= smallest;
}

static void
  reduceRGBMap _IMPL((disp), RGBDisplayParams disp)
{ 
    RGBImage rgb = (RGBImage) disp->generic.image;
    ExternalParams *xparams = ImExternalParams((GenericDisplayParams)disp);
    use_RGBImage(rgb,rgb);
    int depth = DefaultDepthOfScreen(xparams->screen);
    unsigned long *pixel_counts;
    unsigned short pixel_array[32768];
    unsigned long count, midpoint;
    int x, y, num_pixels, allocated;
    struct color_area *areas, *largest_area, *smallest_area;
    struct color_area *new_area, *old_area;
    int n = 1<<depth;
    short srclinelen = 3 * rgb_width;
    byte *srcline, *sp;
    
    if ( !rgb->rgb.pixel_counts ) {
	rgb->rgb.pixel_counts =
	  (unsigned long *) malloc(32768 * sizeof(unsigned long));
    }
    pixel_counts = rgb->rgb.pixel_counts;
    
    if (n > 32768) /* max # of colors we can handle */
      n= 32768;
    
    /* create a histogram of particular pixel occurrances
     */
    
    bzero(pixel_counts, 32768 * sizeof(unsigned long));
    
    srcline = rgb_data;
    for (y = 0; y < rgb_height; y++) {
	for (x = 0, sp = srcline; x < rgb_width; x++) {
	    byte r = *sp++, g = *sp++, b = *sp++;
	    pixel_counts[TRUE3_TO_15BIT(r,g,b)]++;
	}
	srcline += srclinelen;
    }
    
    /* create array of 15-bit pixel values that actually occur in the image
     */
    
    num_pixels= 0;
    for (x= 0; x < 32768; x++) {
	if (pixel_counts[x] > 0)
	  pixel_array[num_pixels++]= (short)x;
    }
    
    /* create color area array and initialize first element
     */
    
    areas= (struct color_area *) alloca(n * sizeof(struct color_area));
    areas[0].pixels= pixel_array;
    areas[0].num_pixels= num_pixels;
    largest_area= smallest_area= (struct color_area *)NULL;
    insertColorArea(pixel_counts, &largest_area, &smallest_area, areas);
    allocated= 1;
    
    /* keep splitting the color area until we have as many color areas as we
     * need
     */
    
    while (allocated < n) {
	
	/* if our largest area can't be broken down, we can't even get the
	 * number of colors they asked us to
	 */
	
	if (largest_area->num_pixels < 2)
	  break;
	
	/* find midpoint of largest area and do split
	 */
	
	qsort(largest_area->pixels, largest_area->num_pixels, sizeof(short),
	      largest_area->sort_func);
	count= 0;
	midpoint= largest_area->pixel_count / 2;
	for (x= 0; x < largest_area->num_pixels; x++) {
	    count += pixel_counts[largest_area->pixels[x]];
	    if (count > midpoint)
	      break;
	}
	if (x == 0) /* degenerate case; divide in half */
	  x= 1;
	new_area= areas + allocated;
	new_area->pixels= largest_area->pixels + x;
	new_area->num_pixels= largest_area->num_pixels - x;
	largest_area->num_pixels= x;
	old_area= largest_area;
	largest_area= largest_area->next;
	if (largest_area)
	  largest_area->prev= (struct color_area *)NULL;
	else
	  smallest_area= (struct color_area *)NULL;
	
	/* recalculate for each area of split and insert in the area list
	 */
	
	insertColorArea(pixel_counts, &largest_area, &smallest_area, old_area);
	insertColorArea(pixel_counts, &largest_area, &smallest_area, new_area);
	
	allocated++;
    }
    
    /* get destination image
     */
    
    /* calculate RGB table from each color area.  this should really calculate
     * a new color by weighting the intensities by the number of pixels, but
     * it's a pain to scale so this just averages all the intensities.  it
     * works pretty well regardless.
     */
    
    if ( rgb_cmap->size < allocated ) {
	freeRGBMapData(rgb_cmap);
	newRGBMapData(rgb_cmap, allocated);
    }
    
    for (x= 0; x < allocated; x++) {
	long red, green, blue, count, pixel;
	
	red= green= blue= 0;
	count= areas[x].pixel_count;
	for (y= 0; y < areas[x].num_pixels; y++) {
	    pixel= areas[x].pixels[y];
	    red += RED_INTENSITY(pixel);
	    green += GREEN_INTENSITY(pixel);
	    blue += BLUE_INTENSITY(pixel);
	    pixel_counts[pixel]= x;
	}
	red /= areas[x].num_pixels;
	green /= areas[x].num_pixels;
	blue /= areas[x].num_pixels;
	
	rgb_cmap->red[x] = (unsigned short)(red << 11);
	rgb_cmap->green[x] = (unsigned short)(green << 11);
	rgb_cmap->blue[x] = (unsigned short)(blue << 11);
    };
    rgb_cmap->used = allocated;
    rgb_cmap->compressed = 1;
}

static byte *
  reduceImage _IMPL((disp, im, im_width, sx, sy, width, height, depth, llen),
		    RGBDisplayParams disp _AND
		    byte *im _AND int im_width _AND
		    int sx _AND int sy _AND int width _AND int height _AND
		    int depth _AND int *llen )
{ 
    RGBImage rgb = (RGBImage) disp->generic.image;
    use_RGBImage(rgb,rgb);
    unsigned long *pixel_counts = rgb->rgb.pixel_counts;
    int x, y;
    byte perpxl = 8/depth;
    short srclinelen = 3 * im_width;
    byte *srcline, *sp;
    short destlinelen;
    byte *dest, *destptr, *destline;

    /* copy old image into new image
     */
    
    *llen = destlinelen = (width+perpxl-1)/perpxl;
    dest = destline = (byte *) calloc( height, destlinelen );

    srcline = im + sy*srclinelen + 3*sx;
    for (y = 0; y < height; y++) {
	sp = srcline;
	destptr = destline;
	for (x = 0; x < width; x++) {
	    byte r = *sp++, g = *sp++, b = *sp++;
	    *destptr++ = disp->rgb.index[pixel_counts[TRUE3_TO_15BIT(r,g,b)]];
	}
	srcline += srclinelen;
	destline += destlinelen;
    }

    return dest;
}
