/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: notifier.c,v 0.9.1.4 1991/12/17 17:57:01 leei Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/wait.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xos.h>

#include <X11/Shell.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>

#include "proto.h"

/*
 * Data structure for application resources.
 */

typedef struct {
    int maxWidth;
    Boolean bell;
    Boolean center;
} AppData, *AppDataPtr;

AppData app_data;

/* resource list */
static XtResource resources[] = {
    { "maxWidth", "MaxWidth", XtRInt, sizeof(int),
      XtOffsetOf(AppData, maxWidth), XtRImmediate, (XtPointer) 640},
    { "bell", "Bell", "Boolean", sizeof(Boolean),
      XtOffsetOf(AppData, bell), XtRString, "True"},
    { "center", "Center", "Boolean", sizeof(Boolean),
      XtOffsetOf(AppData, center), XtRString, "True"}
};

/* Command line option list. */
static XrmOptionDescRec options[] = {
    { "-maxWidth", "*maxWidth", XrmoptionSepArg, NULL },
    { "-bell",     "*bell",     XrmoptionNoArg, "true" },
    { "-nobell",   "*bell",     XrmoptionNoArg, "false" },
    { "-center",   "*center",   XrmoptionNoArg, "true" },
    { "-nocenter", "*center",   XrmoptionNoArg, "false" }
};

/*
 *	Widget and non-widget resources if the application defaults
 *	file can't be found.
 *	[ Generated automatically from Notifier.ad. ]
 */
  
static String fallbackResources[] = {
#include "Notifier.ad.h"
	NULL
};

Boolean stdin_open = True;
Boolean popped_up = False;

static char *notif_str = "";
static int   notif_len = 0;

static void
Clear _IMPL((w, client_data, call_data),
	    Widget w _AND XtPointer client_data _AND XtPointer call_data )
{
    Widget text = (Widget) client_data;
    if ( notif_str && *notif_str ) free(notif_str);
    notif_str = ""; notif_len = 0;
    XtVaSetValues(text, 
		  XtNstring, notif_str, XtNlength, notif_len,
		  XtNinsertPosition, 0, NULL);
}

static void
Hide _IMPL((w, client_data, call_data),
	   Widget w _AND XtPointer client_data _AND XtPointer call_data )
{
    if ( stdin_open ) {
	XtPopdown((Widget)client_data);
	XSync(XtDisplay(w),False);
	popped_up = False;
    } else {
	exit(0);
    }
}

static void 
  maybeWiden _DCL((shell, text, str, len),
		  Widget shell _AND Widget text _AND
		  String str _AND int len);
static void
  CenterWidgetUnderPointer _DCL((w), Widget w);

Boolean 
  WaitForXOrStdin _DCL((dpy, use_stdin), 
		       Display *dpy _AND Boolean use_stdin);
Boolean 
  WaitForStdinOrTimeout _DCL((secs), float secs);

int
main _IMPL((argc,argv), 
	   Cardinal argc _AND String *argv)
{
    XtAppContext app_context;
    Widget toplevel, shell, form, text, hide;
    Display *display;
    int flags;
    Boolean initialized = False;

    /* Make sure that input on stdin is non-blocking. */
    toplevel = XtVaAppInitialize( &app_context, "Notifier",
				  options, XtNumber(options),
				  &argc, argv,
				  fallbackResources,
				  NULL );
    XtGetApplicationResources(toplevel, &app_data, 
			      resources, XtNumber(resources),
			      argv, argc);

    shell = XtVaCreatePopupShell("shell", transientShellWidgetClass, 
				 toplevel, NULL);

    {
	Widget title, clear;
	String label;

	form = 
	  XtVaCreateManagedWidget("form", formWidgetClass, shell, 
				  NULL);

	title =
	  XtVaCreateManagedWidget( "title", labelWidgetClass, form,
				  NULL );
	
	/* If title label is "title" then set to program name. */
	XtVaGetValues(title, XtNlabel, &label, NULL);
	if ( !strcmp("title", label) )
	  XtVaSetValues(title, XtNlabel, XtName(toplevel), NULL);

	text =
	  XtVaCreateManagedWidget( "text", asciiTextWidgetClass, form,
				  XtNtype, XawAsciiString,
				  XtNuseStringInPlace, True,
				  XtNstring, notif_str,
				  XtNlength, notif_len,
				  NULL);

	clear = 
	  XtVaCreateManagedWidget( "clearButton", commandWidgetClass, form,
				  NULL);
	XtAddCallback( clear, XtNcallback, Clear, text);

	hide = 
	  XtVaCreateManagedWidget( "hideButton", commandWidgetClass, form,
				  NULL);
	XtAddCallback( hide, XtNcallback, Hide, shell);
    }

    XtRealizeWidget(shell);
    display = XtDisplay(shell);

    /* Make sure stdin is set up for non-blocking input. */
    flags = fcntl( 0, F_GETFL, 0 );
    fcntl( 0, F_SETFL, flags | FNDELAY );

    while ( True ) {
	XEvent event;

	if ( WaitForXOrStdin(display,stdin_open) ) {
	    int prev_len = notif_len, n; 
	    char buf[BUFSIZ];

	    /* Copy data from stdin to Text widget. */
	    while ( WaitForStdinOrTimeout(0.25) && 
		    (n = read(0, buf, BUFSIZ)) > 0 ) {
		char *newstr = malloc(notif_len+n+1);
		if ( notif_str && *notif_str ) {
		    bcopy(notif_str, newstr, notif_len);
		    free(notif_str);
		}
		bcopy(buf, newstr+notif_len, n);
		newstr[notif_len+n] = '\0';
		notif_str = newstr; notif_len += n;
	    }

	    if ( notif_len != prev_len ) {
		XawTextDisableRedisplay(text);
		XtVaSetValues( text, XtNstring, notif_str, XtNlength, notif_len, NULL );
		if ( !initialized ) {
		    maybeWiden( shell, text, notif_str, notif_len );
		    if ( app_data.center ) CenterWidgetUnderPointer(shell);
		    initialized = True;
		}
		XtVaSetValues(text, XtNinsertPosition, notif_len, NULL);
		XawTextSetSelection(text,prev_len,notif_len);
		XawTextEnableRedisplay(text);
		if ( !popped_up ) XtPopup(shell,XtGrabNone);
		popped_up = True;
	    } else {
		if ( !popped_up ) exit(0);
		XtVaSetValues( hide, XtNlabel, "Quit", NULL );
		stdin_open = False;
	    }
	    if (app_data.bell) XBell(display,50);
	}

	while ( XtAppPending(app_context) ) {
	    XtAppNextEvent(app_context,&event);
	    XtDispatchEvent(&event);
	}
    }
}

#define MIN(a,b)	((a)<(b)? (a): (b))

static void
maybeWiden _IMPL((shell, text, str, len),
		 Widget shell _AND Widget text _AND
		 String str _AND int len)
{
    Dimension twidth;
    Position lm, rm;
    int width = 0;
    XFontStruct *fnt;
    char *p, *q;
    
    /* Get current size of text widget and its vertical scrollbar. */
    XtVaGetValues(text, XtNwidth,&twidth, 
		  XtNleftMargin,&lm, XtNrightMargin,&rm, 
		  XtNfont,&fnt, NULL);

    /* Find the width of the longest line. */
    for ( p = str; p && *p; p = *q? q+1: q ) {
	int w;
	(q = strchr(p,'\n')) || (q = strchr(p,'\0'));
	w = XTextWidth(fnt, p, q-p);
	if ( w > width ) width = w;
    }

    /* 
     * If we need more space increment shell width by 
     * amount required.
     */
    width += lm + rm;
    if ( width > twidth ) {
	Dimension shwidth;
	XtVaGetValues(shell, XtNwidth, &shwidth, NULL);
	shwidth = MIN(app_data.maxWidth, shwidth + width - twidth);
	XtVaSetValues(shell, XtNwidth, shwidth, NULL);
    }
}

/* Centre the widget over the pointer. */
static void
CenterWidgetUnderPointer _IMPL((w), Widget w)
{
    Dimension width, height;
    Window r, c;
    int rx, ry, cx, cy;
    Position x, y;
    unsigned int m;
    
    XtVaGetValues(w, XtNwidth, &width, XtNheight, &height, NULL); 
    XQueryPointer(XtDisplay(w), XtWindow(w), &r,&c, &rx,&ry,&cx,&cy, &m);
    x = rx-width/2;  if ( x < 0 ) x = 0;
    y = ry-height/2; if ( y < 0 ) y = 0;
    XtVaSetValues(w, XtNx, x, XtNy, y, NULL);
}

/* 
 * Blocks for input to either stdin or X connection. 
 * Returns True if stdin is available.
 */
Boolean
WaitForXOrStdin _IMPL((dpy, use_stdin), 
		      Display *dpy _AND Boolean use_stdin)
{
    fd_set r_mask;
    int result;
    int xconn = ConnectionNumber(dpy);
    extern int	errno;

    FD_ZERO(&r_mask);
    do {
        if ( use_stdin ) FD_SET(0, &r_mask); 	/* stdin */
        FD_SET(xconn, &r_mask);		/* display to X */
        result = select(xconn + 1, &r_mask, NULL, NULL, NULL);
        if (result == -1 && errno != EINTR) _XIOError(dpy);
    } while (result <= 0);
    return FD_ISSET(0,&r_mask);
}

/* 
 * Blocks for input to either stdin or X connection. 
 * Returns True if stdin is available.
 */
Boolean
WaitForStdinOrTimeout _IMPL((secs), float secs)
{
    fd_set r_mask;
    int result;

    FD_ZERO(&r_mask);
    do {
	struct timeval timev;
	timev.tv_sec = secs;
	timev.tv_usec = (secs - timev.tv_sec)*1e6;

        FD_SET(0, &r_mask);
        result = select(1, &r_mask, NULL, NULL, &timev);
	if ( !result ) return False;
    } while (result < 0);
    return True;
}
